#include <omp.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include <stdlib.h>
#include "mex.h"
#include "lapack.h"
#include "common.h"
#include "ndsyev.h"
#include "parse_input.h"

#define	ASCENDING_ORDER	1
#define	ABSOLUTE_VALUE		2

void mexFunction(	int nlhs,	mxArray *plhs[],
						int nrhs,	const mxArray *prhs[])
{
	/***********************
	 * [Q,S] = ndsyev(A)
	 *
	 * parallelised eigenvalue decomposition of N-D symmetric matrix
	 * A is an m x m x ... array
	 * in each page of A, an m x m real, single, symmetric matrix is stored
	 * ndsyev finds the eigendecomposition of each page of A
	 * and stores it in a big-ass array
	 * order is an optional parameter to specify how the eigenvalues should
	 * be ordered. four values are permitted:
	 *
	 * s-			descending order
	 * s+			ascending order
	 * u-			descending order of magnitude
	 * u+			ascending order of magnitude
	 *
	 * default order is descending order
	 */

	size_t n_dims_A, n_pages_A;
	size_t M;
	int n, i;
	const mwSize *mws_A;
	const mxArray *mx_A;
	const mxArray *mx_order;
	const float *f_A;
	char sOrder[3];
	unsigned order;

	mxArray *mx_Q, *mx_S;
	float *f_Q, *f_S;
	mwSize *mws_S;

	char jobz, uplo;
	float *f_work;
	int info, lwork;
	float **fp_eigs;
	int *p_reidx;

	int (*comparator)(const void*, const void*);

	/***************
	 * check right number of arguments have been supplied
	 */
	if(nrhs < 1 || nrhs > 2)
	{
		mexErrMsgTxt("Usage: [Q,S] = ndssyev(A, order)");
		return;
	}
	if(nlhs > 2)
	{
		mexErrMsgTxt("too many output arguments");
		return;
	}

	/***************
	 * check arguments have been supplied in the right format
	 */
	mx_A		= prhs[0];
	mx_order	= NULL;
	if(nrhs == 2)
		mx_order = prhs[1];

	/* A must have very particular dimensions */
	if(!mxIsClass(mx_A, "single") || mxIsComplex(mx_A))
	{	mexErrMsgTxt("A must be a real, single array"); }
	n_dims_A	= mxGetNumberOfDimensions(mx_A);
	mws_A		= mxGetDimensions(mx_A);
	M			= mws_A[0];
	if(mws_A[1] != M)
	{	mexErrMsgTxt("A must be an M x M x ... real single array");	}
	n_pages_A= mxGetN(mx_A) / M;
	f_A		= (const float*)mxGetData(mx_A);

	/* order string */
	order		= 0;
	if(mx_order)
	{
		if(mxGetString(mx_order, sOrder, 3) == 0)
		{
			if(strcmp(sOrder, "s-") == 0)				{ order	= 0;	}
			else if(strcmp(sOrder, "s+") == 0)		{ order	= 1;	}	
			else if(strcmp(sOrder, "u-") == 0)		{ order	= 2;	}
			else if(strcmp(sOrder, "u+") == 0)		{ order	= 3;	}
		}
		else
		{
			mexWarnMsgTxt("order should be one of the following string values: s+, s-, u+ or u-");
		}
	}

	/* select comparator */
	switch(order)
	{
	case 1:
		comparator		= compare_float_ptr_ascend;
		break;
	case 2:
		comparator		= compare_float_ptr_abs_descend;
		break;
	case 3:
		comparator		= compare_float_ptr_abs_ascend;
		break;
	default:
		comparator		= compare_float_ptr_descend;
		break;
	}

	/******
	 * allocate memory for output
	 */
	mws_S		= (mwSize*)malloc(n_dims_A * sizeof(mwSize));
	mws_S[0] = M;
	for(n = 0; n < n_dims_A-2; ++n)
		mws_S[1+n] = mws_A[2+n];

	mx_Q		= mxDuplicateArray(mx_A);
	mx_S		= mxCreateNumericArray(n_dims_A-1, mws_S, mxSINGLE_CLASS, mxREAL);
	f_Q		= (float*)mxGetData(mx_Q);
	f_S		= (float*)mxGetData(mx_S);
	free(mws_S);

	/*****
	 * compute eigenvectors/eigenvalues
	 */
	#pragma omp parallel \
		private(f_work, fp_eigs, p_reidx, info, jobz, uplo, n, i, lwork) \
		shared(M, f_Q, f_S, n_pages_A, nlhs, comparator) \
		default(none) \
		num_threads(omp_get_max_threads())
	{
		/* allocate memory for this thread */
		jobz		= (nlhs == 1) ? 'N' : 'V';
		uplo		= 'U';
		lwork		= 4 * M * M;

		f_work	= (float*)malloc(lwork*sizeof(float));
		fp_eigs	= (float**) malloc(M * sizeof(float*));
		p_reidx  = (int*) malloc(M * sizeof(int));

		/* enter parallel for loop */
		#pragma omp for schedule(static) nowait
		for(n = 0; n < n_pages_A; ++n)
		{
			/* get eigenvectors/values */
			//#pragma omp critical
			{
				ssyev(&jobz, &uplo, &M,
						&f_Q[SUB2IND_2D(0,n,M*M)], &M, 
						&f_S[SUB2IND_2D(0,n,M)], f_work, &lwork, &info);
			}
			/*
			if(info < 0)
				mexErrMsgTxt("ssyev failed with illegal value");
			if(info > 0)
				mexErrMsgTxt("ssyev failed to converge");
			*/

			/* sort 
			 * sort an array of pointers, but comparison is done on dereferenced value
			 * then subtract original index to get reordered index
			 */
			for(i = 0; i < M; ++i)
				fp_eigs[i] = &f_S[SUB2IND_2D(i,n,M)];
			qsort(fp_eigs, M, sizeof(float*), comparator);
			for(i = 0; i < M; ++i)
				p_reidx[i] = fp_eigs[i] - &f_S[SUB2IND_2D(0,n,M)];

			/* reorder eigenvalues */
			memcpy(f_work, &f_S[SUB2IND_2D(0,n,M)], sizeof(float)*M);
			for(i = 0; i < M; ++i)
				f_S[SUB2IND_2D(i,n,M)] = f_work[p_reidx[i]];

			/* reorder eigenvectors, if necessary */
			if(nlhs == 2)
			{
				memcpy(f_work, &f_Q[SUB2IND_3D(0,0,n,M,M)], sizeof(float)*M*M);
				for(i = 0; i < M; ++i)
				{
					memcpy(	&f_Q[SUB2IND_3D(0,i,n,M,M)], 
								&f_work[SUB2IND_2D(0,p_reidx[i],M)], sizeof(float)*M);
				}
			}
		}

		/* free memory for this thread */
		free(f_work);
		free(fp_eigs);
		free(p_reidx);
	}

	/****
	 * allocate output
	 */
	if(nlhs < 2)	{ mxDestroyArray(mx_Q); }
	else				{ plhs[1] = mx_Q;			}
	
	if(nlhs < 1)	{ mxDestroyArray(mx_S); }
	else				{ plhs[0] = mx_S;			}
}


int compare_float_ptr_abs_ascend(const void *a, const void *b)
{
	if ( fabs(**(float**)a) <  fabs(**(float**)b) ) return -1;
	if (     (**(float**)a) ==     (**(float**)b) ) return 0;
	return 1;
}

int compare_float_ptr_abs_descend(const void *a, const void *b)
{
	if ( fabs(**(float**)a) <  fabs(**(float**)b) ) return 1;
	if (     (**(float**)a) ==     (**(float**)b) ) return 0;
	return -1;
}

int compare_float_ptr_ascend(const void *a, const void *b)
{
	if ( **(float**)a <  **(float**)b ) return -1;
	if ( **(float**)a == **(float**)b ) return 0;
	return 1;
}

int compare_float_ptr_descend(const void *a, const void *b)
{
	if ( **(float**)a <  **(float**)b ) return 1;
	if ( **(float**)a == **(float**)b ) return 0;
	return -1;
}

